#include <Servo.h>
#include <LiquidCrystal_I2C.h>
// Popescu Ioana-Denisa 336CA

// Pin definitions
int led = 11;      // Yellow Led
int button = 7;   // Button
int led_red = 13;
int led_green = 12;

// LCD Pins
const int trigPin = 9;
const int echoPin = 10;

// Servo variables
Servo myServo;
int servoPin = 8;
volatile int Time = 0;
volatile float distance;  // Sensor distance
static bool forward = true;   // Running direction
volatile int pos = 0;         // Servo position
volatile bool move = false;
int servoSpeed = 5;           // Speed of servo movement

// LCD variables
LiquidCrystal_I2C lcd(0x27, 16, 2);
unsigned long lcdUpdateTime = 0;     // Time of the last LCD update
unsigned long lcdUpdateInterval = 900;  // Delay between LCD updates (in milliseconds)

// Button variables
static bool buttonPressed = false;   // Tracks if the button was previously pressed

// Function prototypes
void updateLCD();
void moveServo();
void nonBlockingServoControl();
void ADC_Potentiometer();

void setup() {
  Serial.begin(9600);

  // Initialize ultrasonic sensor pins
  pinMode(trigPin, OUTPUT);
  pinMode(echoPin, INPUT);

  // Initialize LCD
  lcd.init();
  lcd.backlight();
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Automatic Cat");
  lcd.setCursor(0, 1);
  lcd.print("Toy");

  // Initialize LED pins
  DDRB |= (1 << PB3);
  pinMode(led_green, OUTPUT);
  pinMode(led_red, OUTPUT);
  digitalWrite(led_green, HIGH);
  digitalWrite(led_red, LOW);

  // Enable Pin Change Interrupt 2 for the button
  PCICR |= B00000100;
  PCMSK2 |= B10000000;

  // Initialize serial communication and servo
  Serial.begin(9600);
  myServo.attach(5);

  // Configure Timer0 for PWM
  TCCR0A = _BV(COM0B1) | _BV(WGM00) | _BV(WGM01);  // Set Fast PWM mode and Clear OC0B on Compare Match
  TCCR0B = _BV(WGM02) | _BV(CS02) | _BV(CS00);     // Set prescaler to 8
  OCR0A = 31;                                      // Set TOP value to achieve a frequency of 50 Hz
  OCR0B = 15;                                      // Set initial servo position
  TIMSK0 |= _BV(OCIE0A);                           // Enable Timer 0 compare match A interrupt

  sei();                                           // Enable interrupts

}

void loop() {
  int status = PORTB & (1 << PB3);

  if (!status) {
    ADC_Potentiometer();
  } else {
    // Read distance from ultrasonic sensor
    digitalWrite(trigPin, LOW);
    delayMicroseconds(2);
    digitalWrite(trigPin, HIGH);
    delayMicroseconds(10);
    digitalWrite(trigPin, LOW);
    long duration = pulseIn(echoPin, HIGH);
    distance = duration * 0.034 / 2;

    if (move && distance < 10) { // Move the Servo randomly if the detection is close enough
      nonBlockingServoControl();
    }
    updateLCD();
  }
}

// Function to update the LCD display
void updateLCD() {
  if (millis() - lcdUpdateTime >= lcdUpdateInterval) {
    lcdUpdateTime = millis();
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Distance: ");
    lcd.print(distance);
  }
}

// Function to move the servo to the desired position
void moveServo() {
  myServo.write(pos);
}

// Function for non-blocking servo control using Timer0 interrupt
void nonBlockingServoControl() {
  static unsigned long prevTime = 0;
  unsigned long currentTime = millis();

  if (currentTime - prevTime >= servoSpeed) {
    prevTime = currentTime;
    moveServo();
  }
}

// Interrupt Service Routine for Timer0 compare match A interrupt
ISR(TIMER0_COMPA_vect) {
  Time++;

  if (Time == 20) {
    if (forward) {
      pos += 5;               // Increment position
      if (pos >= 180) {
        forward = false;      // Reached the maximum position, change direction
      }
    } else {
      pos -= 5;               // Decrement position
      if (pos <= 0) {
        forward = true;       // Reached the starting position, change direction
      }
    }
    move = true;
    Time = 0;
  }
}

// Interrupt Service Routine for button press
ISR(PCINT2_vect) {
  bool buttonStatus = digitalRead(button);  // Read the current button status

  if (buttonStatus && !buttonPressed) {
    // Button was just pressed
    digitalWrite(led, !digitalRead(led));  // Toggle the LED status
    buttonPressed = true;                  // Update buttonPressed flag

    // Disable Timer0 interrupt during button press
    TIMSK0 &= ~_BV(OCIE0A);
  } else if (!buttonStatus && buttonPressed) {
    // Button was just released
    buttonPressed = false;                  // Reset buttonPressed flag

    // Enable Timer0 interrupt after button release
    TIMSK0 |= _BV(OCIE0A);
  }
}

// Function to read and control the servo position using a potentiometer
void ADC_Potentiometer() {
  ADMUX |= B01000000; // Reference Voltage
  ADCSRA |= B01000111; // Prescaler + Enable
  ADCSRA |= (1 << ADSC);

  while (!(ADCSRA & (1 << ADIF)));
  uint16_t result = ADC;

  if (result < 500) {
    digitalWrite(led_red, LOW);
    digitalWrite(led_green, HIGH);
  } else {
    digitalWrite(led_red, HIGH);
    digitalWrite(led_green, LOW);
  }

  if (result < 1010) {
    int servoPosition = map(result, 0, 1023, 0, 180);
    myServo.write(servoPosition);
    Serial.println(result);
  }
}
